@@ -41,6 +41,7 @@
 #endif
 #include <nm-device-infiniband.h>
 #include <nm-device-bond.h>
+#include <nm-device-bridge.h>
 #include <nm-device-vlan.h>
 #include <nm-utils.h>
 #include <nm-setting-ip4-config.h>
@@ -90,14 +91,17 @@ static NmcOutputField nmc_fields_dev_list_sections[] = {
 	{"DHCP4",             N_("DHCP4"),             0, NULL, 0},  /* 8 */
 	{"IP6",               N_("IP6"),               0, NULL, 0},  /* 9 */
 	{"DHCP6",             N_("DHCP6"),             0, NULL, 0},  /* 10 */
+	{"BOND",              N_("BOND"),              0, NULL, 0},  /* 11 */
+	{"VLAN",              N_("VLAN"),              0, NULL, 0},  /* 12 */
+	{"CONNECTIONS",       N_("CONNECTIONS"),       0, NULL, 0},  /* 13 */
 	{NULL,                NULL,                    0, NULL, 0}
 };
 #if WITH_WIMAX
-#define NMC_FIELDS_DEV_LIST_SECTIONS_ALL     "GENERAL,CAPABILITIES,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,WIMAX-PROPERTIES,NSP,IP4,DHCP4,IP6,DHCP6"
-#define NMC_FIELDS_DEV_LIST_SECTIONS_COMMON  "GENERAL,CAPABILITIES,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,WIMAX-PROPERTIES,NSP,IP4,DHCP4,IP6,DHCP6"
+#define NMC_FIELDS_DEV_LIST_SECTIONS_ALL     "GENERAL,CAPABILITIES,BOND,VLAN,CONNECTIONS,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,WIMAX-PROPERTIES,NSP,IP4,DHCP4,IP6,DHCP6"
+#define NMC_FIELDS_DEV_LIST_SECTIONS_COMMON  "GENERAL,CAPABILITIES,BOND,VLAN,CONNECTIONS,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,WIMAX-PROPERTIES,NSP,IP4,DHCP4,IP6,DHCP6"
 #else
-#define NMC_FIELDS_DEV_LIST_SECTIONS_ALL     "GENERAL,CAPABILITIES,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,IP4,DHCP4,IP6,DHCP6"
-#define NMC_FIELDS_DEV_LIST_SECTIONS_COMMON  "GENERAL,CAPABILITIES,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,IP4,DHCP4,IP6,DHCP6"
+#define NMC_FIELDS_DEV_LIST_SECTIONS_ALL     "GENERAL,CAPABILITIES,BOND,VLAN,CONNECTIONS,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,IP4,DHCP4,IP6,DHCP6"
+#define NMC_FIELDS_DEV_LIST_SECTIONS_COMMON  "GENERAL,CAPABILITIES,BOND,VLAN,CONNECTIONS,WIFI-PROPERTIES,AP,WIRED-PROPERTIES,IP4,DHCP4,IP6,DHCP6"
 #endif
 
 /* Available fields for 'dev list' - GENERAL part */
@@ -121,9 +125,20 @@ static NmcOutputField nmc_fields_dev_list_general[] = {
 	{"CONNECTION",        N_("CONNECTION"),        51, NULL, 0},  /* 16 */
 	{NULL, NULL, 0, NULL, 0}
 };
-#define NMC_FIELDS_DEV_LIST_GENERAL_ALL     "NAME,DEVICE,TYPE,VENDOR,PRODUCT,DRIVER,DRIVER-VERSION,FIRMWARE-VERSION,HWADDR,STATE,REASON,UDI,IP-IFACE,NM-MANAGED,AUTOCONNECT,FIRMWARE-MISSING,CONNECTION"
+#define NMC_FIELDS_DEV_LIST_GENERAL_ALL     "NAME,DEVICE,TYPE,VENDOR,PRODUCT,DRIVER,DRIVER-VERSION,FIRMWARE-VERSION,HWADDR,STATE,REASON,UDI,IP-IFACE,"\
+                                            "NM-MANAGED,AUTOCONNECT,FIRMWARE-MISSING,CONNECTION"
 #define NMC_FIELDS_DEV_LIST_GENERAL_COMMON  "NAME,DEVICE,TYPE,VENDOR,PRODUCT,DRIVER,HWADDR,STATE"
 
+/* Available fields for 'dev list' - CONNECTIONS part */
+static NmcOutputField nmc_fields_dev_list_connections[] = {
+	{"NAME",                       N_("NAME"),                       10, NULL, 0},  /* 0 */
+	{"AVAILABLE-CONNECTION-PATHS", N_("AVAILABLE-CONNECTION-PATHS"), 80, NULL, 0},  /* 1 */
+	{"AVAILABLE-CONNECTIONS",      N_("AVAILABLE-CONNECTIONS"),      80, NULL, 0},  /* 2 */
+	{NULL, NULL, 0, NULL, 0}
+};
+#define NMC_FIELDS_DEV_LIST_CONNECTIONS_ALL     "AVAILABLE-CONNECTION-PATHS,AVAILABLE-CONNECTIONS"
+#define NMC_FIELDS_DEV_LIST_CONNECTIONS_COMMON  "AVAILABLE-CONNECTION-PATHS,AVAILABLE-CONNECTIONS"
+
 /* Available fields for 'dev list' - CAPABILITIES part */
 static NmcOutputField nmc_fields_dev_list_cap[] = {
 	{"NAME",            N_("NAME"),            13, NULL, 0},  /* 0 */
@@ -152,10 +167,12 @@ static NmcOutputField nmc_fields_dev_list_wifi_prop[] = {
 	{"WPA2",       N_("WPA2"),         6, NULL, 0},  /* 3 */
 	{"TKIP",       N_("TKIP"),         6, NULL, 0},  /* 4 */
 	{"CCMP",       N_("CCMP"),         6, NULL, 0},  /* 5 */
+	{"AP",         N_("AP"),           6, NULL, 0},  /* 6 */
+	{"ADHOC",      N_("ADHOC"),        6, NULL, 0},  /* 7 */
 	{NULL,         NULL,               0, NULL, 0}
 };
-#define NMC_FIELDS_DEV_LIST_WIFI_PROP_ALL     "NAME,WEP,WPA,WPA2,TKIP,CCMP"
-#define NMC_FIELDS_DEV_LIST_WIFI_PROP_COMMON  "NAME,WEP,WPA,WPA2,TKIP,CCMP"
+#define NMC_FIELDS_DEV_LIST_WIFI_PROP_ALL     "NAME,WEP,WPA,WPA2,TKIP,CCMP,AP,ADHOC"
+#define NMC_FIELDS_DEV_LIST_WIFI_PROP_COMMON  "NAME,WEP,WPA,WPA2,TKIP,CCMP,AP,ADHOC"
 
 #if WITH_WIMAX
 /* Available fields for 'dev list' - wimax properties part */
@@ -210,6 +227,24 @@ static NmcOutputField nmc_fields_dev_wimax_list[] = {
 #define NMC_FIELDS_DEV_WIMAX_LIST_FOR_DEV_LIST  "NAME,"NMC_FIELDS_DEV_WIMAX_LIST_COMMON
 #endif
 
+/* Available fields for 'dev list' - BOND part */
+static NmcOutputField nmc_fields_dev_list_bond_prop[] = {
+	{"NAME",            N_("NAME"),     18, NULL, 0},  /* 0 */
+	{"SLAVES",          N_("SLAVES"),   20, NULL, 0},  /* 1 */
+	{NULL,              NULL,            0, NULL, 0}
+};
+#define NMC_FIELDS_DEV_LIST_BOND_PROP_ALL     "NAME,SLAVES"
+#define NMC_FIELDS_DEV_LIST_BOND_PROP_COMMON  "NAME,SLAVES"
+
+/* Available fields for 'dev list' - VLAN part */
+static NmcOutputField nmc_fields_dev_list_vlan_prop[] = {
+	{"NAME",            N_("NAME"),     18, NULL, 0},  /* 0 */
+	{"ID",              N_("ID"),        5, NULL, 0},  /* 1 */
+	{NULL,              NULL,            0, NULL, 0}
+};
+#define NMC_FIELDS_DEV_LIST_VLAN_PROP_ALL     "NAME,ID"
+#define NMC_FIELDS_DEV_LIST_VLAN_PROP_COMMON  "NAME,ID"
+
 
 /* glib main loop variable - defined in nmcli.c */
 extern GMainLoop *loop;
@@ -220,7 +255,7 @@ static void
 usage (void)
 {
 	fprintf (stderr,
-	         _("Usage: nmcli dev { COMMAND | help }\n\n"
+	         _("Usage: nmcli device { COMMAND | help }\n\n"
 #if WITH_WIMAX
 	         "  COMMAND := { status | list | disconnect | wifi | wimax }\n\n"
 #else
@@ -233,8 +268,9 @@ usage (void)
 	         "  wifi connect <(B)SSID> [password <password>] [wep-key-type key|phrase] [iface <iface>] [bssid <BSSID>] [name <name>]\n"
 	         "               [--private] [--nowait] [--timeout <timeout>]\n"
 #if WITH_WIMAX
-	         "  wimax [list [iface <iface>] [nsp <name>]]\n\n"
+	         "  wimax [list [iface <iface>] [nsp <name>]]\n"
 #endif
+	         "\n"
 	         ));
 }
 
@@ -286,6 +322,8 @@ device_type_to_string (NMDevice *device)
 		return NM_SETTING_BOND_SETTING_NAME;
 	case NM_DEVICE_TYPE_VLAN:
 		return NM_SETTING_VLAN_SETTING_NAME;
+	case NM_DEVICE_TYPE_BRIDGE:
+		return NM_SETTING_BRIDGE_SETTING_NAME;
 	default:
 		return _("Unknown");
 	}
@@ -403,11 +441,13 @@ detail_access_point (gpointer data, gpointer user_data)
 	if (security_str->len > 0)
 		g_string_truncate (security_str, security_str->len-1);  /* Chop off last space */
 
-	ap_name = g_strdup_printf ("AP%d", info->index++); /* AP */
+	ap_name = g_strdup_printf ("AP[%d]", info->index++); /* AP */
 	info->nmc->allowed_fields[0].value = ap_name;
 	info->nmc->allowed_fields[1].value = ssid_str;
 	info->nmc->allowed_fields[2].value = bssid;
-	info->nmc->allowed_fields[3].value = mode == NM_802_11_MODE_ADHOC ? _("Ad-Hoc") : mode == NM_802_11_MODE_INFRA ? _("Infrastructure") : _("Unknown");
+	info->nmc->allowed_fields[3].value = mode == NM_802_11_MODE_ADHOC ? _("Ad-Hoc")
+	                                   : mode == NM_802_11_MODE_INFRA ? _("Infrastructure")
+	                                   : _("Unknown");
 	info->nmc->allowed_fields[4].value = freq_str;
 	info->nmc->allowed_fields[5].value = bitrate_str;
 	info->nmc->allowed_fields[6].value = strength_str;
@@ -461,7 +501,7 @@ detail_wimax_nsp (NMWimaxNsp *nsp, NmCli *nmc, NMDevice *dev, int idx)
 	}
 
 	quality_str = g_strdup_printf ("%u", nm_wimax_nsp_get_signal_quality (nsp));
-	nsp_name = g_strdup_printf ("NSP%d", idx); /* NSP */
+	nsp_name = g_strdup_printf ("NSP[%d]", idx); /* NSP */
 
 	nmc->allowed_fields[0].value = nsp_name;
 	nmc->allowed_fields[1].value = nm_wimax_nsp_get_name (nsp);
@@ -567,6 +607,8 @@ show_device_info (gpointer data, gpointer user_data)
 				hwaddr = nm_device_bond_get_hw_address (NM_DEVICE_BOND (device));
 			else if (NM_IS_DEVICE_VLAN (device))
 				hwaddr = nm_device_vlan_get_hw_address (NM_DEVICE_VLAN (device));
+			else if (NM_IS_DEVICE_BRIDGE (device))
+				hwaddr = nm_device_bridge_get_hw_address (NM_DEVICE_BRIDGE (device));
 
 			state_str = g_strdup_printf ("%d (%s)", state, nmc_device_state_to_string (state));
 			reason_str = g_strdup_printf ("%d (%s)", reason, nmc_device_reason_to_string (reason));
@@ -627,6 +669,7 @@ show_device_info (gpointer data, gpointer user_data)
 			was_output = TRUE;
 		}
 
+
 		/* Wireless specific information */
 		if ((NM_IS_DEVICE_WIFI (device))) {
 			NMDeviceWifiCapabilities wcaps;
@@ -649,6 +692,8 @@ show_device_info (gpointer data, gpointer user_data)
 				nmc->allowed_fields[3].value = (wcaps & NM_WIFI_DEVICE_CAP_RSN) ? _("yes") : _("no");
 				nmc->allowed_fields[4].value = (wcaps & NM_WIFI_DEVICE_CAP_CIPHER_TKIP) ? _("yes") : _("no");
 				nmc->allowed_fields[5].value = (wcaps & NM_WIFI_DEVICE_CAP_CIPHER_CCMP) ? _("yes") : _("no");
+				nmc->allowed_fields[6].value = (wcaps & NM_WIFI_DEVICE_CAP_AP) ? _("yes") : _("no");
+				nmc->allowed_fields[7].value = (wcaps & NM_WIFI_DEVICE_CAP_ADHOC) ? _("yes") : _("no");
 
 				nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_SECTION_PREFIX;
 				print_fields (nmc->print_fields, nmc->allowed_fields); /* Print values */
@@ -790,6 +835,112 @@ show_device_info (gpointer data, gpointer user_data)
 			if (dhcp6 && !strcasecmp (nmc_fields_dev_list_sections[section_idx].name, nmc_fields_dev_list_sections[10].name))
 				was_output = print_dhcp6_config (dhcp6, nmc, nmc_fields_dev_list_sections[10].name);
 		}
+
+		/* Bond-specific information */
+		if ((NM_IS_DEVICE_BOND (device))) {
+			if (!strcasecmp (nmc_fields_dev_list_sections[section_idx].name, nmc_fields_dev_list_sections[11].name)) {
+				const GPtrArray *slaves;
+				GString *bond_slaves_str;
+				int idx;
+
+				bond_slaves_str = g_string_new (NULL);
+				slaves = nm_device_bond_get_slaves (NM_DEVICE_BOND (device));
+				for (idx = 0; slaves && idx < slaves->len; idx++) {
+					NMDevice *slave = g_ptr_array_index (slaves, idx);
+					const char *iface = nm_device_get_iface (slave);
+
+					if (iface) {
+						g_string_append (bond_slaves_str, iface);
+						g_string_append_c (bond_slaves_str, ' ');
+					}
+				}
+				if (bond_slaves_str->len > 0)
+					g_string_truncate (bond_slaves_str, bond_slaves_str->len-1);  /* Chop off last space */
+
+				nmc->allowed_fields = nmc_fields_dev_list_bond_prop;
+				nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_FIELD_NAMES;
+				nmc->print_fields.indices = parse_output_fields (NMC_FIELDS_DEV_LIST_BOND_PROP_ALL, nmc->allowed_fields, NULL);
+				print_fields (nmc->print_fields, nmc->allowed_fields); /* Print header */
+
+				nmc->allowed_fields[0].value = nmc_fields_dev_list_sections[11].name;  /* "BOND" */
+				nmc->allowed_fields[1].value = bond_slaves_str->str;
+
+				nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_SECTION_PREFIX;
+				print_fields (nmc->print_fields, nmc->allowed_fields); /* Print values */
+
+				g_string_free (bond_slaves_str, TRUE);
+				was_output = TRUE;
+			}
+		}
+
+		/* VLAN-specific information */
+		if ((NM_IS_DEVICE_VLAN (device))) {
+			if (!strcasecmp (nmc_fields_dev_list_sections[section_idx].name, nmc_fields_dev_list_sections[12].name)) {
+				char * vlan_id_str = g_strdup_printf ("%u", nm_device_vlan_get_vlan_id (NM_DEVICE_VLAN (device)));
+
+				nmc->allowed_fields = nmc_fields_dev_list_vlan_prop;
+				nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_FIELD_NAMES;
+				nmc->print_fields.indices = parse_output_fields (NMC_FIELDS_DEV_LIST_VLAN_PROP_ALL, nmc->allowed_fields, NULL);
+				print_fields (nmc->print_fields, nmc->allowed_fields); /* Print header */
+
+				nmc->allowed_fields[0].value = nmc_fields_dev_list_sections[12].name;  /* "VLAN" */
+				nmc->allowed_fields[1].value = vlan_id_str;
+
+				nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_SECTION_PREFIX;
+				print_fields (nmc->print_fields, nmc->allowed_fields); /* Print values */
+
+				g_free (vlan_id_str);
+				was_output = TRUE;
+			}
+		}
+
+		/* section CONNECTIONS */
+		if (!strcasecmp (nmc_fields_dev_list_sections[section_idx].name, nmc_fields_dev_list_sections[13].name)) {
+			const GPtrArray *avail_cons;
+			GString *ac_paths_str;
+			char **ac_arr = NULL;
+			int i;
+
+			nmc->allowed_fields = nmc_fields_dev_list_connections;
+			nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_FIELD_NAMES;
+			nmc->print_fields.indices = parse_output_fields (NMC_FIELDS_DEV_LIST_CONNECTIONS_ALL, nmc->allowed_fields, NULL);
+			print_fields (nmc->print_fields, nmc->allowed_fields); /* Print header */
+
+			/* available-connections */
+			avail_cons = nm_device_get_available_connections (device);
+			ac_paths_str = g_string_new (NULL);
+			if (avail_cons->len) {
+				ac_arr = g_new (char *, avail_cons->len + 1);
+				ac_arr[avail_cons->len] = NULL;
+			}
+			for (i = 0; avail_cons && (i < avail_cons->len); i++) {
+				NMRemoteConnection *avail_con = g_ptr_array_index (avail_cons, i);
+				const char *ac_path = nm_connection_get_path (NM_CONNECTION (avail_con));
+				const char *ac_id = nm_connection_get_id (NM_CONNECTION (avail_con));
+				const char *ac_uuid = nm_connection_get_uuid (NM_CONNECTION (avail_con));
+
+				ac_arr[i] = g_strdup_printf ("%s | %s", ac_uuid, ac_id);
+
+				if (i == 0)
+					g_string_printf (ac_paths_str, "%s/{", NM_DBUS_PATH_SETTINGS);
+				else
+					g_string_append_c (ac_paths_str, ',');
+				g_string_append (ac_paths_str, strrchr (ac_path, '/') + 1);
+			}
+			if (ac_paths_str->len > 0)
+				g_string_append_c (ac_paths_str, '}');
+
+			set_val_str (nmc->allowed_fields, 0, nmc_fields_dev_list_sections[13].name);  /* "CONNECTIONS" */
+			set_val_str (nmc->allowed_fields, 1, ac_paths_str->str);
+			set_val_arr (nmc->allowed_fields, 2, (const char **) ac_arr);
+
+			nmc->print_fields.flags = multiline_flag | mode_flag | escape_flag | NMC_PF_FLAG_SECTION_PREFIX;
+			print_fields (nmc->print_fields, nmc->allowed_fields); /* Print values */
+
+			g_string_free (ac_paths_str, TRUE);
+			g_strfreev (ac_arr);
+			was_output = TRUE;
+		}
 	}
 
 	if (sections_array)
@@ -895,7 +1046,7 @@ do_devices_list (NmCli *nmc, int argc, char **argv)
 			iface_specified = TRUE;
 
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: '%s' argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: '%s' argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -932,7 +1083,7 @@ do_devices_list (NmCli *nmc, int argc, char **argv)
 			NMDevice *candidate = g_ptr_array_index (devices, i);
 			const char *dev_iface = nm_device_get_iface (candidate);
 
-			if (!strcmp (dev_iface, iface))
+			if (!g_strcmp0 (dev_iface, iface))
 				device = candidate;
 		}
 		if (!device) {
@@ -1038,7 +1189,7 @@ do_device_disconnect (NmCli *nmc, int argc, char **argv)
 			iface_specified = TRUE;
 
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1048,7 +1199,7 @@ do_device_disconnect (NmCli *nmc, int argc, char **argv)
 			wait = FALSE;
 		} else if (strcmp (*argv, "--timeout") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1096,7 +1247,7 @@ do_device_disconnect (NmCli *nmc, int argc, char **argv)
 		NMDevice *candidate = g_ptr_array_index (devices, i);
 		const char *dev_iface = nm_device_get_iface (candidate);
 
-		if (!strcmp (dev_iface, iface))
+		if (!g_strcmp0 (dev_iface, iface))
 			device = candidate;
 	}
 
@@ -1166,7 +1317,7 @@ do_device_wifi_list (NmCli *nmc, int argc, char **argv)
 	while (argc > 0) {
 		if (strcmp (*argv, "iface") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1174,7 +1325,7 @@ do_device_wifi_list (NmCli *nmc, int argc, char **argv)
 		} else if (strcmp (*argv, "bssid") == 0 || strcmp (*argv, "hwaddr") == 0) {
 			/* hwaddr is deprecated and will be removed later */
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1235,7 +1386,7 @@ do_device_wifi_list (NmCli *nmc, int argc, char **argv)
 			NMDevice *candidate = g_ptr_array_index (devices, i);
 			const char *dev_iface = nm_device_get_iface (candidate);
 
-			if (!strcmp (dev_iface, iface)) {
+			if (!g_strcmp0 (dev_iface, iface)) {
 				device = candidate;
 				break;
 			}
@@ -1431,7 +1582,7 @@ find_wifi_device_by_iface (const GPtrArray *devices, const char *iface, int *idx
 
 		if (iface) {
 			/* If a iface was specified then use it. */
-			if (strcmp (dev_iface, iface) == 0) {
+			if (g_strcmp0 (dev_iface, iface) == 0) {
 				device = candidate;
 				break;
 			}
@@ -1538,14 +1689,14 @@ do_device_wifi_connect_network (NmCli *nmc, int argc, char **argv)
 	while (argc > 0) {
 		if (strcmp (*argv, "iface") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
 			iface = *argv;
 		} else if (strcmp (*argv, "bssid") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1559,14 +1710,14 @@ do_device_wifi_connect_network (NmCli *nmc, int argc, char **argv)
 			}
 		} else if (strcmp (*argv, "password") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
 			password = *argv;
 		} else if (strcmp (*argv, "wep-key-type") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1583,7 +1734,7 @@ do_device_wifi_connect_network (NmCli *nmc, int argc, char **argv)
 			}
 		} else if (strcmp (*argv, "name") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1594,7 +1745,7 @@ do_device_wifi_connect_network (NmCli *nmc, int argc, char **argv)
 			wait = FALSE;
 		} else if (strcmp (*argv, "--timeout") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1815,14 +1966,14 @@ do_device_wimax_list (NmCli *nmc, int argc, char **argv)
 	while (argc > 0) {
 		if (strcmp (*argv, "iface") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
 			iface = *argv;
 		} else if (strcmp (*argv, "nsp") == 0) {
 			if (next_arg (&argc, &argv) != 0) {
-				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *argv);
+				g_string_printf (nmc->return_text, _("Error: %s argument is missing."), *(argv-1));
 				nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 				goto error;
 			}
@@ -1882,7 +2033,7 @@ do_device_wimax_list (NmCli *nmc, int argc, char **argv)
 			NMDevice *candidate = g_ptr_array_index (devices, i);
 			const char *dev_iface = nm_device_get_iface (candidate);
 
-			if (!strcmp (dev_iface, iface)) {
+			if (!g_strcmp0 (dev_iface, iface)) {
 				device = candidate;
 				break;
 			}
@@ -2024,10 +2175,13 @@ do_devices (NmCli *nmc, int argc, char **argv)
 			nmc->return_value = do_device_wimax (nmc, argc-1, argv+1);
 		}
 #endif
-		else if (strcmp (*argv, "help") == 0) {
+		else if (   matches (*argv, "help") == 0
+		         || (g_str_has_prefix (*argv, "-") &&  matches ((*argv)+1, "help") == 0)
+		         || (g_str_has_prefix (*argv, "--") && matches ((*argv)+2, "help") == 0)) {
 			usage ();
 		}
 		else {
+			usage ();
 			g_string_printf (nmc->return_text, _("Error: 'dev' command '%s' is not valid."), *argv);
 			nmc->return_value = NMC_RESULT_ERROR_USER_INPUT;
 		}
